MODULE cmor_getinput

IMPLICIT NONE
SAVE
PRIVATE
PUBLIC cmor_getDate, cmor_getTime, similar, read_table

CONTAINS 


!! JY Stuff

  FUNCTION cmor_getDate(datesep)
    !
    ! Get the current date
    !
    ! Parameters
    ! - datesep : separator used in the date
    !     default: "/"
    !
    ! Result
    !
    ! Return a 10-character string in the form mm/dd/yyyy

    IMPLICIT NONE
                                                                   
    ! Function result
    CHARACTER(len=10) :: cmor_getDate
                                                  
    CHARACTER(len=1), INTENT(in), OPTIONAL :: datesep
    
    CHARACTER(len=1) :: sep
    CHARACTER(len=8) :: rtndate

                                                             
    ! Get the separator
    IF (PRESENT(datesep)) THEN
        sep = datesep
    ELSE
        sep = "/"
    END IF
                                                              
    CALL DATE_AND_TIME(date=rtndate)
                                                                 
    cmor_getDate = rtndate(5:6) // sep // rtndate(7:8) // sep // rtndate(1:4)
                                                                  
  END FUNCTION cmor_getDate

                                                            
  FUNCTION cmor_getTime(timesep)
    !
    ! Get the current time
    !
    ! Parameters
    ! - timesep : separator used in the time
    !     default: ":"
    !
    !
    ! Result
    !
    ! Return a 8-character string in the form hh:mm:ss

    IMPLICIT NONE

    ! Function result
    CHARACTER(len=8) :: cmor_getTime
                                                                   
    CHARACTER(len=1), INTENT(in), OPTIONAL :: timesep
                                                        
    CHARACTER(len=1) :: sep
    CHARACTER(len=10) :: rtntime
                                                   
    ! Get the separator
    IF (PRESENT(timesep)) THEN
        sep = timesep
    ELSE
        sep = ":"
    END IF
                                                        
    CALL DATE_AND_TIME(time=rtntime)

    cmor_getTime = rtntime(1:2) // sep // rtntime(3:4) // sep // rtntime(5:6)

  END FUNCTION cmor_getTime


  FUNCTION similar(str1, str2)

    ! compares two strings after removing leading and trailing blanks and 
    ! ignoring case.  

    IMPLICIT NONE

    LOGICAL :: similar
    CHARACTER(len=*), INTENT(IN) :: str1
    CHARACTER(len=*), INTENT(IN) :: str2

    CHARACTER(len=LEN(str1)) :: low1
    CHARACTER(len=LEN(str2)) :: low2
    INTEGER :: len1, i

    low1 = ADJUSTL(str1)
    low2 = ADJUSTL(str2)

    len1 = LEN_TRIM(low1)
    similar = .TRUE.
    IF (len1 /= LEN_TRIM(low2)) THEN
       similar = .FALSE.
    ELSE
       ! convert strings to lower case
       DO i=1,len1
          IF ( LGE(low1(i:i),'A') .AND. LLE(low1(i:i),'Z') ) &
               low1(i:i) = ACHAR( IACHAR(low1(i:i)) + 32 )
          IF ( LGE(low2(i:i),'A') .AND. LLE(low2(i:i),'Z') ) &
               low2(i:i) = ACHAR( IACHAR(low2(i:i)) + 32 )
       END DO

       IF (TRIM(low1) /= TRIM(low2)) similar=.FALSE.
    END IF

    RETURN
  END FUNCTION similar

  FUNCTION read_table(filename)

    use cmor_internal_tables
    IMPLICIT NONE

    INTEGER :: read_table
    CHARACTER(len=*), INTENT(IN) :: filename

    CHARACTER(len=a32) :: keyword
    CHARACTER(len=a512) :: ljust_string
    INTEGER :: ierr, iaxis, ivar, i, ntables, naxes, nvars, iaxis1, ivar1, j
    INTEGER :: istat, idim
    CHARACTER(len=a16), SAVE :: section   ! identifies section of table

    PRINT*, ' '
    PRINT*, 'CMOR is reading a MIP input table: ', TRIM(filename)
  
    read_table = 0
    ntables = num_mip_tables
    nvars = num_vars
    naxes = num_axes

    iaxis1=naxes+1
    ivar1 = nvars+1

    ! open file containing table
    OPEN(UNIT=idev_mip_table, FILE=TRIM(filename), STATUS='OLD',   &
         ACTION='READ', IOSTAT=ierr)

    IF (ierr /= 0) THEN
       PRINT*, 'Error encountered in CMOR: read_table'
       PRINT*, 'Error opening MIP Table in file: ', TRIM(filename)
       STOP
    END IF

    section='preamble'

    read_lines: DO

       CALL get_input(idev_mip_table, ierr, keyword, ljust_string)
       IF (ierr /= 0) THEN
          IF (ierr == -1) THEN
             
             PRINT*, '   CMOR is exiting loop over input table lines'

             EXIT
          END IF
          PRINT*, 'Error encountered in CMOR: get_input'
          PRINT*, 'which was called from read_table'
          STOP
       END IF

       SELECT CASE (keyword)
       CASE ('variable_entry')
          section = 'variable'
          nvars = nvars+1
          IF (nvars>max_variables) THEN
             PRINT*, 'Error encountered in CMOR: read_table'
             PRINT*, 'Number of variables in the MIP table exceeds max_variables'
             STOP
          END IF
          var(nvars)%table_entry = TRIM(ljust_string)
          var(nvars)%arg_cntl = var_arg_cntl
          CYCLE
       CASE ('axis_entry')
          section = 'axis'
          naxes = naxes+1
          IF (naxes>max_axes) THEN
             PRINT*, 'Error encountered in CMOR: read_table'
             PRINT*, 'Number of axes in the MIP table exceeds max_axes'   
             STOP
          END IF
          axis(naxes)%table_entry = TRIM(ljust_string)
          axis(naxes)%arg_cntl = axis_arg_cntl
          CYCLE
       END SELECT

       SELECT CASE (section)

       CASE ('variable')
          ierr = enter_variable_info(nvars, keyword, TRIM(ljust_string))

       CASE ('axis') 
          ierr = enter_axis_info(naxes, keyword, TRIM(ljust_string))

       CASE ('preamble')
          ierr = enter_preamble_info(ntables, keyword, TRIM(ljust_string))
          
       END SELECT
       IF (ierr /= 0) THEN
          PRINT*, 'Error encountered in CMOR: enter_' // TRIM(section) // '_info'
          PRINT*, 'which was called from read_table'
          STOP
       END IF
    END DO read_lines

    ! close file
    CLOSE(idev_mip_table, IOSTAT=ierr)

    IF (ierr /= 0) THEN
       read_table=ierr
       PRINT*, 'Error encountered in CMOR: read_table'
       PRINT*, 'Error attempting to close file: ', idev_mip_table
       STOP
    END IF

    ! loop through axes and fill in other axis table information
    ! also enter singleton (scalar) dimensions in dimension table
    iaxis = iaxis1
    DO WHILE (iaxis<=naxes) 
       axis(iaxis)%mip_table = mip_tbl(ntables)%table_id
       IF (LEN_TRIM(axis(iaxis)%out_name) == 0)  &
            axis(iaxis)%out_name = axis(iaxis)%table_entry

       !  if bounds required define bounds attribute

       IF ((axis(iaxis)%arg_cntl(6) == 1) .OR.    &
            (axis(iaxis)%arg_cntl(6) == 2) .OR.   &
            (associated(axis(iaxis)%zfac_bnds_names))) THEN
          axis(iaxis)%bounds_attrib = TRIM(axis(iaxis)%out_name) // '_bnds'
       END IF
       IF ((.not. similar(axis(iaxis)%type, 'character')) .and.  &
            associated(axis(iaxis)%requested_labels)) THEN
          ALLOCATE(axis(iaxis)%requested_vals(    &
               SIZE(axis(iaxis)%requested_labels)), STAT=ierr)
          IF (ierr /= 0) THEN
             PRINT*, 'Error encountered in CMOR: read_table'
             PRINT*, 'error allocating axis(iaxis)%requested_vals'
             STOP
          END IF
          DO i=1,SIZE(axis(iaxis)%requested_labels)
             READ(axis(iaxis)%requested_labels(i), '(F32.1)')    &
                  axis(iaxis)%requested_vals(i)
             axis(iaxis)%requested_labels(i) = ''
          END DO
       END IF
       IF (LEN_TRIM(axis(iaxis)%singleton_label) /= 0) THEN
          num_dims = num_dims + 1
          IF (num_dims > max_dimensions) THEN
             PRINT*, 'Error encountered in CMOR: read_table'
             PRINT*, 'num_dims exceeds max_dimensions'
             STOP
          END IF
          idim = num_dims
          adim(idim)%axis_idx = iaxis
          adim(idim)%lendim = 1

          IF (similar(axis(iaxis)%type, 'character')) THEN

             ALLOCATE(adim(idim)%labels(1), STAT=istat)
             IF (istat /= 0) THEN
                PRINT*, 'Error encountered in CMOR: read_table'
                PRINT*, 'error allocating adim(idim)%values'
                STOP
             END IF

             adim(idim)%labels(1) = axis(iaxis)%singleton_label

          ELSE
            
             adim(idim)%users_units = axis(iaxis)%out_units

             ALLOCATE(adim(idim)%values(1), STAT=istat)
             IF (istat /= 0) THEN
                PRINT*, 'Error encountered in CMOR: read_table'
                PRINT*, 'error allocating adim(idim)%values'
                STOP
             END IF

             READ(axis(iaxis)%singleton_label, '(F32.1)')    &
                  adim(idim)%values(1)

          ENDIF

          IF (axis(iaxis)%singleton_bnds(1) /= init_undef_double) THEN
             ALLOCATE(adim(idim)%bounds(2,1), STAT=istat)
             IF (istat /= 0) THEN
                PRINT*, 'Error encountered in CMOR: read_table'
                PRINT*, 'error allocating adim(idim)%bounds'
                STOP
             END IF
             adim(idim)%bounds(:,1) = axis(iaxis)%singleton_bnds(:)
          END IF
       ENDIF

       iaxis = iaxis + 1
    END DO
    num_axes = naxes

    ! loop through variables and fill in other variable table information
    ivar = ivar1
    DO while (ivar<=nvars)
       var(ivar)%mip_table = mip_tbl(ntables)%table_id
       IF (LEN_TRIM(var(ivar)%out_name) == 0)  &
            var(ivar)%out_name = var(ivar)%table_entry
       var(ivar)%rank = 0
       DO j=1,max_rank
          IF (LEN_TRIM(var(ivar)%axis_names(j)) == 0) EXIT
          var(ivar)%rank = var(ivar)%rank + 1
          var(ivar)%axis_idxs(j)=0
          IF (INDEX(var(ivar)%axis_names(j), 'zlevel') /= 0) THEN
             var(ivar)%axis_idxs(j) = -1
          ELSE
             do i=iaxis1,naxes
                IF (similar(var(ivar)%axis_names(j), axis(i)%table_entry)) THEN
                   var(ivar)%axis_idxs(j)=i
                   EXIT
                END IF
             END DO
          END IF
       END DO
       ivar = ivar + 1
    END DO
    num_vars = nvars

    PRINT*, 'CMOR successfully read the MIP table.'
    PRINT*, ' '

    RETURN
  END FUNCTION read_table





  FUNCTION enter_axis_info(iaxis, keyword, string_in)

    USE cmor_internal_tables
    IMPLICIT NONE

    INTEGER :: enter_axis_info
    INTEGER, INTENT(IN) :: iaxis
    CHARACTER(len=*), INTENT(IN) :: keyword
    CHARACTER(len=*), INTENT(IN) :: string_in

    CHARACTER(len=a32), DIMENSION(max_requested_coords) :: list
    DOUBLE PRECISION, ALLOCATABLE, DIMENSION(:,:) :: bnds_copy

    INTEGER, DIMENSION(2) :: ishape
    INTEGER :: i, j, ierr, jold

    enter_axis_info=0

    SELECT CASE (keyword)

    CASE ('standard_name')
       axis(iaxis)%standard_name = string_in
    CASE ('units')
       axis(iaxis)%out_units = string_in
    CASE ('axis')
       axis(iaxis)%axis = string_in
    CASE ('positive')
       axis(iaxis)%positive_attrib = string_in
    CASE ('long_name')
       axis(iaxis)%long_name = string_in
    CASE ('stored_direction')
       IF (string_in == 'decreasing') THEN
          axis(iaxis)%out_direction = -1
       ELSE IF (string_in == 'increasing') THEN
          axis(iaxis)%out_direction = 1
       ELSE
          PRINT*, 'Error encountered in CMOR: enter_axis_info'
          PRINT*, 'stored_direction must be either "increasing" or' // &
               ' "decreasing", but was ' // TRIM(string_in) // '"'
          enter_axis_info = -1
          STOP
       END IF
    CASE ('valid_min')
       READ(string_in, *) axis(iaxis)%valid_min
    CASE ('valid_max')
       READ(string_in, *) axis(iaxis)%valid_max
    CASE ('type')
       axis(iaxis)%type = string_in
    CASE ('out_name')
       axis(iaxis)%out_name = string_in
    CASE ('coords_attrib')
       axis(iaxis)%coords_attrib = string_in
    CASE ('tol_on_requests')
       READ(string_in, *) axis(iaxis)%tol_on_requests
    CASE ('value')
       ! READ(string_in, *) axis(iaxis)%singleton_val
       READ(string_in, *) axis(iaxis)%singleton_label
    CASE ('bounds_values')
       READ(string_in, *) (axis(iaxis)%singleton_bnds(i), i=1,2)
    CASE ('requested')

       IF (associated(axis(iaxis)%requested_labels)) THEN
          list(1:SIZE(axis(iaxis)%requested_labels)) =   &
               axis(iaxis)%requested_labels
          DEALLOCATE(axis(iaxis)%requested_labels)
       ELSE
          list = ''
       END IF

       CALL parser(string_in, list, j)

       ALLOCATE(axis(iaxis)%requested_labels(j), STAT=ierr)
       IF (ierr /= 0) THEN
          PRINT*, 'Error encountered in CMOR: enter_axis_info'
          PRINT*, 'error allocating axis(iaxis)%requested_labels'
          enter_axis_info=-1
          STOP  
       END IF
       axis(iaxis)%requested_labels = list(1:j)

    CASE ('bounds_requested')

       IF (associated(axis(iaxis)%requested_bounds)) THEN
          ishape(1) = SIZE(axis(iaxis)%requested_bounds, 1)
          ishape(2) = SIZE(axis(iaxis)%requested_bounds, 2)
          ALLOCATE(bnds_copy(ishape(1), ishape(2)), STAT=ierr)
          IF (ierr /= 0) THEN
             PRINT*, 'Error encountered in CMOR: enter_axis_info'
             PRINT*, 'error allocating bnds_copy'
             STOP  
          END IF
          bnds_copy = axis(iaxis)%requested_bounds
          DEALLOCATE(axis(iaxis)%requested_bounds)
       END IF
       list = ''
       CALL parser(string_in, list, j)

       jold = 0
       IF (ALLOCATED(bnds_copy)) jold = SIZE(bnds_copy,2)
       ALLOCATE(axis(iaxis)%requested_bounds(2,j+jold), STAT=ierr)
       IF (ierr /= 0) THEN
          PRINT*, 'Error encountered in CMOR: enter_axis_info'
          PRINT*, 'error allocating axis(iaxis)%requested_bounds'
          STOP  
       END IF

       IF (ALLOCATED(bnds_copy)) THEN 
          axis(iaxis)%requested_bounds(1:2,1:jold) = bnds_copy
          DEALLOCATE(bnds_copy)
       END IF

       DO i=jold+1,jold+j
          IF (LEN_TRIM(list(i)) == 0) EXIT
          READ(list(i-jold), '(F32.1)')    &
               axis(iaxis)%requested_bounds(MOD((i-1),2)+1, (i-1)/2+1)
       END DO

    CASE ('formula')
       axis(iaxis)%formula = string_in
    CASE ('convert_to')
       axis(iaxis)%convert_to = string_in
    CASE ('z_factors')
       IF (associated(axis(iaxis)%zfac_names)) THEN
          list(1:SIZE(axis(iaxis)%zfac_names)) =   &
               axis(iaxis)%zfac_names
          DEALLOCATE(axis(iaxis)%zfac_names)
       ELSE
          list = ''
       END IF

       CALL parser(string_in, list, j)

       ALLOCATE(axis(iaxis)%zfac_names(j), STAT=ierr)
       IF (ierr /= 0) THEN
          PRINT*, 'Error encountered in CMOR: enter_axis_info'
          PRINT*, 'error allocating axis(iaxis)%zfac_names'
          STOP  
       END IF
       axis(iaxis)%zfac_names = list(1:j)
    CASE ('z_bounds_factors')
       IF (associated(axis(iaxis)%zfac_bnds_names)) THEN
          list(1:SIZE(axis(iaxis)%zfac_bnds_names)) =   &
               axis(iaxis)%zfac_bnds_names
          DEALLOCATE(axis(iaxis)%zfac_bnds_names)
       ELSE
          list = ''
       END IF

       CALL parser(string_in, list, j)

       ALLOCATE(axis(iaxis)%zfac_bnds_names(j), STAT=ierr)
       IF (ierr /= 0) THEN
          PRINT*, 'Error encountered in CMOR: enter_axis_info'
          PRINT*, 'error allocating axis(iaxis)%zfac_bnds_names'
          STOP  
       END IF
       axis(iaxis)%zfac_bnds_names = list(1:j)
    CASE ('required')
       CALL arg_assign('required', string_in, axis_args, axis(iaxis)%arg_cntl)
    CASE ('indeterminate')
       CALL arg_assign('indeterminate', string_in, axis_args,    &
            axis(iaxis)%arg_cntl)
    CASE ('optional')
       CALL arg_assign('optional', string_in, axis_args, axis(iaxis)%arg_cntl)
    CASE ('ignored')
       CALL arg_assign('ignored', string_in, axis_args, axis(iaxis)%arg_cntl)
    CASE ('forbidden')
       CALL arg_assign('forbidden', string_in, axis_args,  &
            axis(iaxis)%arg_cntl)

    CASE DEFAULT

       ! error exit

    END SELECT

    RETURN
  END FUNCTION enter_axis_info



  FUNCTION enter_variable_info(ivar, keyword, string_in)

    USE cmor_internal_tables
    IMPLICIT NONE

    INTEGER :: enter_variable_info
    INTEGER, INTENT(IN) :: ivar
    CHARACTER(len=*), INTENT(IN) :: keyword
    CHARACTER(len=*), INTENT(IN) :: string_in

    INTEGER :: i, j, ierr

    enter_variable_info=0

    SELECT CASE (keyword)

    CASE ('standard_name')
       var(ivar)%standard_name = string_in
    CASE ('units')
       var(ivar)%out_units = string_in
    CASE ('cell_methods')
       var(ivar)%cell_methods = string_in
    CASE ('long_name')
       var(ivar)%long_name = string_in
    CASE ('comment')
       var(ivar)%comment = string_in
    CASE ('positive')
       var(ivar)%out_positive = string_in
    CASE ('dimensions')
       CALL parser(string_in, var(ivar)%axis_names)
    CASE ('valid_min')
       READ(string_in, *) var(ivar)%valid_min
    CASE ('valid_max')
       READ(string_in, *) var(ivar)%valid_max
    CASE ('ok_min_mean_abs')
       READ(string_in, *) var(ivar)%ok_min_mean_abs
    CASE ('ok_max_mean_abs')
       READ(string_in, *) var(ivar)%ok_max_mean_abs
    CASE ('type')
       var(ivar)%type = string_in
    CASE ('out_name')
       var(ivar)%out_name = string_in

    CASE ('required')
       CALL arg_assign('required', string_in, var_args, var(ivar)%arg_cntl)
    CASE ('indeterminate')
       CALL arg_assign('indeterminate', string_in, var_args, &
            var(ivar)%arg_cntl)
    CASE ('optional')
       CALL arg_assign('optional', string_in, var_args, var(ivar)%arg_cntl)
    CASE ('ignored')
       CALL arg_assign('ignored', string_in, var_args, var(ivar)%arg_cntl)
    CASE ('forbidden')
       CALL arg_assign('forbidden', string_in, var_args, var(ivar)%arg_cntl)

    CASE DEFAULT

       ! error exit

    END SELECT

    RETURN
  END FUNCTION enter_variable_info




  FUNCTION enter_preamble_info(itable, keyword, string_in)

    USE cmor_internal_tables
    IMPLICIT NONE

    INTEGER :: enter_preamble_info
    INTEGER, INTENT(IN) :: itable
    CHARACTER(len=*), INTENT(IN) :: keyword
    CHARACTER(len=*), INTENT(IN) :: string_in

    REAL :: aa
    CHARACTER(len=8), save :: sub

    enter_preamble_info = 0

    !     required/indeterminate/optional/ignored/forbidden
    SELECT CASE (keyword)

    CASE ('subroutine_entry')

       SELECT CASE (string_in)

       CASE ('cmor_axis')
          sub = 'axis'
       CASE ('cmor_variable')
          sub = 'variable'
       CASE ('cmor_write')
          sub = 'write'
       END SELECT

    CASE ('cmor_version')
       READ(string_in, '(f5.2)') aa
       IF (aa > cmor_version) THEN
          PRINT*, 'Error encountered in CMOR: enter_preamble_info'
          PRINT*, 'You should obtain the latest version of CMOR before '
          PRINT*, 'processing this MIP table.  Go to:'
          PRINT*, 'http://www-pcmdi.llnl.gov/cmip/cmor_readme.html'
          enter_preamble_info=-1
          STOP
       END IF
    CASE ('cf_version')
       mip_tbl(itable)%cf_version = string_in
    CASE ('project_id')
       mip_tbl(itable)%project_id = string_in
    CASE ('table_id')
       mip_tbl(itable)%table_id = string_in
    CASE ('table_date')
       mip_tbl(itable)%table_date = string_in
    CASE ('magic_number')
       READ(string_in,*) mip_tbl(itable)%magic_number
    CASE ('approx_interval')
       READ(string_in,*) mip_tbl(itable)%approx_interval
    CASE ('missing_value')
       READ(string_in,*) mip_tbl(itable)%single_missing
    CASE ('double_missing_value')
       READ(string_in,*) mip_tbl(itable)%double_missing
    CASE ('integer_missing_value')
       READ(string_in,*) mip_tbl(itable)%integer_missing
    CASE ('expt_id_ok') 
       CALL parser(string_in, mip_tbl(itable)%expt_id_ok)
    CASE ('required')
       SELECT CASE (sub)
       CASE ('axis')
          CALL arg_assign('required', string_in, axis_args, axis_arg_cntl)
       CASE ('variable')
          CALL arg_assign('required', string_in, var_args, var_arg_cntl)
       CASE ('write')
          CALL arg_assign('required', string_in, write_args, write_arg_cntl)
       END SELECT
    CASE ('indeterminate')
       SELECT CASE (sub)
       CASE ('axis')
          CALL arg_assign('indeterminate', string_in, axis_args, axis_arg_cntl)
       CASE ('variable')
          CALL arg_assign('indeterminate', string_in, var_args, var_arg_cntl)
       CASE ('write')
          CALL arg_assign('indeterminate', string_in, write_args, write_arg_cntl)
       END SELECT
    CASE ('optional')
       SELECT CASE (sub)
       CASE ('axis')
          CALL arg_assign('optional', string_in, axis_args, axis_arg_cntl)
       CASE ('variable')
          CALL arg_assign('optional', string_in, var_args, var_arg_cntl)
       CASE ('write')
          CALL arg_assign('optional', string_in, write_args, write_arg_cntl)
       END SELECT
    CASE ('ignored')
       SELECT CASE (sub)
       CASE ('axis')
          CALL arg_assign('ignored', string_in, axis_args, axis_arg_cntl)
       CASE ('variable')
          CALL arg_assign('ignored', string_in, var_args, var_arg_cntl)
       CASE ('write')
          CALL arg_assign('ignored', string_in, write_args, write_arg_cntl)
       END SELECT
    CASE ('forbidden')
       SELECT CASE (sub)
       CASE ('axis')
          CALL arg_assign('forbidden', string_in, axis_args, axis_arg_cntl)
       CASE ('variable')
          CALL arg_assign('forbidden', string_in, var_args, var_arg_cntl)
       CASE ('write')
          CALL arg_assign('forbidden', string_in, write_args, write_arg_cntl)
       END SELECT
       
    CASE DEFAULT
       
       PRINT*, 'Error encountered in CMOR: enter_preamble_info'
       PRINT*, 'The following input line contains a keyword that is not recognized by CMOR: '
       PRINT*, 'string_in = ', TRIM(string_in)
       enter_preamble_info=-1
       STOP
       
    END SELECT
               
    RETURN
  END FUNCTION enter_preamble_info


  SUBROUTINE parser(string_in, items, nn)

    IMPLICIT NONE

    CHARACTER(len=*), INTENT(IN)  :: string_in
    CHARACTER(len=*), INTENT(INOUT), DIMENSION(:) :: items
    INTEGER, INTENT(OUT), OPTIONAL :: nn

    ! local variables
    CHARACTER(len=LEN(string_in)) :: str
    INTEGER :: len1, nitems, i, quote1, quote2

    nitems = 0
    DO i = 1, size(items)
       IF (LEN_TRIM(items(i)) <= 0) THEN
          nitems = i-1
          EXIT 
       END IF
    END DO

    str = string_in 
    quote1 = INDEX(str, "'")
    DO WHILE (quote1 > 0)
       str = str(quote1+1:)
       ! find next quote
       quote2 = INDEX(str, "'")
       IF (quote2 <= 0) THEN
          PRINT*, 'Error encountered in CMOR: parser'
          PRINT*, 'The following input line has no closing quote'
          PRINT*, str
          STOP
       END IF
       nitems = nitems + 1
       items(nitems) = str(1:quote2-1)
       str = ADJUSTL(str(quote2+1:))
       quote1 = INDEX(str, "'")
    END DO
    len1 = INDEX((str // ' '), ' ')-1
    DO WHILE (len1>0)
       nitems = nitems + 1
       items(nitems) = str(1:len1)
       str = ADJUSTL(str(len1+1:)) // ' '
       len1 = INDEX(str, ' ')-1
    END DO
       
    IF (PRESENT(nn)) nn = nitems

    RETURN
  END SUBROUTINE parser


  SUBROUTINE arg_assign(keyword, list, sub_args, sub_arg_cntl)

    IMPLICIT NONE

    CHARACTER(len=*), INTENT(IN) :: keyword
    CHARACTER(len=*), INTENT(IN) :: list
    CHARACTER(len=*), DIMENSION(:) :: sub_args
    INTEGER, DIMENSION(:) :: sub_arg_cntl

    CHARACTER(len=LEN(list)) :: str, str1
    INTEGER :: i, j, n

    SELECT CASE (keyword)
    CASE ('required')
       i=1
    CASE ('indeterminate')
       i=2
    CASE ('optional')
       i=3
    CASE ('ignored')
       i=4
    CASE ('forbidden')
       i=5
    END SELECT

    str = ADJUSTL(list)

    DO 
       n=index((str // ' '), ' ')
       IF (n <= 1) EXIT
       str1 = str(1:n-1)
       DO j = 1, SIZE(sub_args, 1)
          IF (TRIM(sub_args(j)) == str1) sub_arg_cntl(j)=i
       END DO
       str = ADJUSTL(str(n:))
    END DO

    RETURN
  END SUBROUTINE arg_assign


  SUBROUTINE get_input(input_device, ierr, keyword, trimmed_string)

    IMPLICIT NONE

    INTEGER, INTENT(IN) :: input_device
    INTEGER, INTENT(OUT) :: ierr
    CHARACTER(len=*), INTENT(OUT) :: keyword
    CHARACTER(len=*), INTENT(OUT) :: trimmed_string


    ! local variables
    CHARACTER(len=LEN(trimmed_string)) :: orig_line
    INTEGER :: len1

    len1=0
    ! Find next line of input, skipping lines with an '!' as the first     
    !      non-blank character and for other lines removing the portion 
    !      following an '!' (including the '!')

    DO WHILE (len1 == 0)
       READ(input_device, '(A)', IOSTAT=ierr) trimmed_string
       ! PRINT*, 'line read: ', TRIM(trimmed_string)
       IF (ierr /= 0) THEN
          !   the 4001 error value is not standard, but required on 
          !    the SGI
          IF ((ierr == -1) .or. (ierr == -4001)) THEN
             ! reached the end of file
             ierr = -1
             PRINT*, '   CMOR reached end of MIP table.' 
             RETURN
          ELSE
             PRINT*, 'Error encountered in CMOR: get_input'
             PRINT*, 'Error reading next line of input from MIP table near:'
             PRINT*, TRIM(trimmed_string)
             PRINT*, 'READ IOSTAT= ', ierr
             STOP
          END IF
       END IF

       trimmed_string = ADJUSTL(trimmed_string)
       orig_line = TRIM(trimmed_string)

       len1 = INDEX(trimmed_string, '!')-1
       IF (len1 > 0) THEN
          trimmed_string=trimmed_string(1:len1)
          len1 = LEN_TRIM(trimmed_string)
       ELSE IF (len1 < 0) THEN
          len1 = LEN_TRIM(trimmed_string)
       END IF

    END DO

    !  copy first item in string (which must be terminated by a ':')
    len1 = INDEX(trimmed_string,':')
    IF (len1>0) THEN
       keyword = trimmed_string(1:len1-1)
    ELSE
       PRINT*, 'Error encountered in CMOR: get_input'
       PRINT*, 'Error in input table:  the first "word" in a non-comment'
       PRINT*, '   line must be a keyword terminated by a ":"'  
       PRINT*, '   This error was found in the following line: '
       PRINT*, '  ' // TRIM(orig_line)
       
       ierr = -2
       STOP
    END IF
    
    trimmed_string = ADJUSTL(trimmed_string(len1+1:))
    
    IF (LEN_TRIM(trimmed_string)==0) THEN
       PRINT*, 'Error encountered in CMOR: get_input'
       PRINT*, 'Error in input table:  Words should follow the keyword "' // &
            TRIM(keyword) // '"'
       PRINT*, '   but nothing (or only a comment) was found'
       PRINT*, '   The error was encountered reading the following line: '
       PRINT*, '  ' // TRIM(orig_line)
       
       ierr = -3
       STOP
    END IF

    keyword = TRIM(keyword)

    ! PRINT*, 'done reading keyword: ', TRIM(keyword)
    ! PRINT*, 'with value: ', TRIM(trimmed_string)
    RETURN
  END SUBROUTINE get_input
  
END MODULE cmor_getinput
